/* $Id: e2_command.c 2167 2006-07-28 03:24:50Z tpgww $
 *
 * Copyright (C) 2004-2006 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/command/e2_command.c
 * @brief functions for handling e2 commands
 *
 **/

#include "e2_command.h"
#include "e2_task.h"
#include <string.h>
#include <sys/wait.h>
#include <signal.h>
#include <pwd.h>
#ifdef E2_ASYNC
#include <pthread.h>
#endif
#include "e2_alias.h"
#include "e2_fs.h"

#define CMDNAMELEN 30

gboolean tochild = FALSE;	//TRUE when sending a command to a child
const gchar *shellcmd;	//the command interpreter to use for external shell
#ifdef E2_ASYNC
extern pthread_mutex_t task_mutex;
#endif

static gint _e2_command_run_sync (gchar *command,
	gchar **args, gboolean show, gchar *prefix);
static gboolean _e2_command_watch (GIOChannel *ioc,
	GIOCondition cond, E2_TaskRuntime *rt, gboolean error);
//static void _e2_command_sigchld_handler (gint num);

/**
 * @brief block incoming SIGCHLD signals
 * @return
 **/
//static
void _e2_command_block_childsignal (void)
{
	sigset_t set;	//, oldset;

	sigemptyset (&set);
	sigaddset (&set, SIGCHLD);
	sigprocmask (SIG_BLOCK, &set, NULL);	//&oldset);
}
/**
 * @brief unblock incoming SIGCHLD signals
 * @return
 **/
//static
void _e2_command_unblock_childsignal (void)
{
	sigset_t set;	//, oldset;

	//re-install the signal handler (some systems need this) WRONG
//	signal (SIGCHLD, _e2_command_sigchld_handler);

	//and unblock it
	sigemptyset (&set);
	sigaddset (&set, SIGCHLD);
	sigprocmask (SIG_UNBLOCK, &set, NULL);	//&oldset);
}
/*
static void signal_handler (gint sig)
{
	printd (DEBUG, "signal: %d", sig);
}
*/
/**
 * @brief clean up data for child processes which are finished
 * Note that opened pipes etc will come here, even though
 * they're not logged as children
 * @param num the signal number (17)
 * @return
 **/
//copied from gtk faq @ http://gtk.org
static void _e2_command_sigchld_handler (gint num)
{
	printd (DEBUG, "sigchld_handler (num:%d)", num);
	pid_t pid;
	gint status;

//not needed SIGCHLD signals blocked while its handler is in progress
//	_e2_command_block_childsignal ();

	//get stopped child(ren), if any
	while ((pid = waitpid ((pid_t) -1, &status, WNOHANG | WUNTRACED)) > 0)
	{
#if DEBUG_LEVEL > 2
		gint exitstatus;
		if (WIFEXITED (status))
		{
			exitstatus = WEXITSTATUS (status);
			printd (DEBUG, "child (%d) exit status = %d", (gint) pid,
				exitstatus);
		}
		else if (WIFSTOPPED (status))
		{
			exitstatus = WSTOPSIG (status);
			printd (DEBUG, "child (%d) stopped by signal %d", (gint) pid,
				exitstatus);
		}
		else if (WIFSIGNALED (status))
		{
			exitstatus = WTERMSIG (status);
			printd (DEBUG, "child (%d) terminated by signal %d", (gint) pid,
				exitstatus);
		}
		else
			printd (DEBUG, "child (%d) exited magically", (gint) pid);
#endif
		// record child exit status if appropriate
		if (!WIFSTOPPED (status))
		{
			E2_TaskRuntime *rt = e2_task_find_task ((glong) pid);
			if (rt != NULL)
			{
				E2_CommandTaskData *qed = (E2_CommandTaskData *)rt->extradata;
				qed->exit = (WIFEXITED (status)) ? WEXITSTATUS (status) : -1;
			}
#if DEBUG_LEVEL > 2
			else
				printd (WARN, "Received signal from unrecorded child process (%d)", (gint) pid);
#endif
		}
	}
	//re-install the signal handler (some systems need this?)
	signal (SIGCHLD, _e2_command_sigchld_handler);

//see comment above
//	_e2_command_unblock_childsignal ();
}
/**
 * @brief set handler for SIGCHLD signals
 * @return
 **/
static void _e2_command_set_sigchild_handler (void)
{
	struct sigaction sigdata;
	sigdata.sa_handler = _e2_command_sigchld_handler;
	sigemptyset (&sigdata.sa_mask);
	sigdata.sa_flags = SA_RESTART;
	sigaction (SIGCHLD, &sigdata, NULL);
/*	signal(SIGCHLD, signal_handler);
	signal(SIGPOLL, signal_handler);
	signal(SIGABRT,signal_handler);
	signal(SIGALRM,signal_handler);
	signal(SIGHUP,signal_handler);
	signal(SIGINT,signal_handler);
	signal(SIGPIPE,signal_handler);
	signal(SIGQUIT,signal_handler);
	signal(SIGTERM,signal_handler);
	signal(SIGUSR1,signal_handler);
	signal(SIGUSR2,signal_handler); */
}
/**
 * @brief process non-error-iochannel data
 * @param ioc the channel passing the data
 * @param cond flags indicating the condition(s) that have been satisfied
 * @param rt pointer to data struct for the command to which the channel applies
 * @return FALSE if the event source should be removed
 **/
static gboolean _e2_command_watch_std (GIOChannel *ioc,
	GIOCondition cond, E2_TaskRuntime *rt)
{
	return _e2_command_watch (ioc, cond, rt, FALSE);
}
/**
 * @brief process error-iochannel data
 * @param ioc the channel passing the data
 * @param cond flags indicating the condition(s) that have been satisfied
 * @param rt pointer to data struct for the command to which the channel applies
 * @return FALSE if the event source should be removed
 **/
static gboolean _e2_command_watch_err (GIOChannel *ioc,
	GIOCondition cond, E2_TaskRuntime *rt)
{
	return _e2_command_watch (ioc, cond, rt, TRUE);
}
/**
 * @brief process iochannel data
 *
 * Data are grabbed in parcels up to 4095 bytes, which are then separated
 * into lines (\n), and those lines are sent to the output printer
 * Will shut the channel down upon error
 *
 * @param ioc the channel event source
 * @param cond flags indicating the condition(s) that have been satisfied
 * @param rt pointer to data struct for the command to which the channel applies
 * @param error TRUE if this call relates to the error channel
 * @return FALSE if the event source should be removed
 **/
static gboolean _e2_command_watch (GIOChannel *ioc,
	GIOCondition cond, E2_TaskRuntime *rt, gboolean error)
{
//	printd (DEBUG, "watch (ioc:,cond:%d,rt:)", cond);
	E2_CommandTaskData *cmddata = (E2_CommandTaskData *)rt->extradata;
	if (cond & (G_IO_IN | G_IO_PRI))
	{
		gchar buf[5000];	//space (rounded) to read 4096 and add trailing \0
		//buf[0] = buf[4096] = '\0';
		gsize len = 0;
		GIOStatus ret = g_io_channel_read_chars (ioc, buf, 4096 * sizeof(gchar), &len, NULL);
		if (ret == G_IO_STATUS_NORMAL && len > 0)	//len>0 is also a proxy for ret == G_IO_STATUS_EOF
		{
			if (cmddata->show || error)
			{
//				if (len < 4096)
					*(buf + len) = '\0';
				if (error)
					e2_output_print (rt->current_tab, buf, rt->pidstr,
						FALSE, E2_ERRORTAGS, NULL);
				else
					e2_output_print (rt->current_tab, buf, rt->pidstr,
						FALSE, NULL);
/* line-by-line processing needed if we want to parse and handle CR's BS's etc
				{
					gchar save;	//ignore compiler warning
					gchar *sep;
					gchar *print = buf;
					gboolean more = TRUE;
					while (more)
					{
						//always ascii \n, don't need utf8 search
						if ((sep = strchr (print, '\n')) != NULL)
						{
							sep++;
							save = *sep;
							*sep = '\0';
						}
						else
							more = FALSE;

						//printd (DEBUG, "printing for tab %x", rt->current_tab);
						if (error)
							e2_output_print (rt->current_tab, print, rt->pidstr,
								FALSE, E2_ERRORTAGS, NULL);
						else
							e2_output_print (rt->current_tab, print, rt->pidstr,
								FALSE, NULL);
						if (more)
						{
							*sep = save;
							print = sep;
						}
					}
				}
*/
			}
			//if (ret == G_IO_STATUS_NORMAL) always TRUE, here
				return TRUE;
		}
	}
	if (cond & (G_IO_ERR | G_IO_HUP | G_IO_NVAL))
	{
		gint fd;
		if (rt->status == E2_TASK_RUNNING)
		{	//first pass (via normal input channel, normally with flag G_IO_HUP)
//see the sigchld_handler for this rt->exit = (cond & (G_IO_ERR | G_IO_NVAL)) ? 1 : 0;
			fd = g_io_channel_unix_get_fd (ioc);
			g_io_channel_shutdown (ioc, TRUE, NULL);
// crasher	g_io_channel_unref (ioc);
			e2_fs_close (fd);
			rt->status = E2_TASK_COMPLETED;
		}
		else
		{	//second pass (via error input channel, normally with flag G_IO_HUP)
			_e2_command_block_childsignal ();
			if (cmddata->show && e2_option_bool_get ("fileop-show"))
			{
				gchar *message = g_strdup_printf (">%s (%ld) %s '%d'\n",
//					rt->command, rt->pid, _("returned"), WEXITSTATUS (rt->exit));
					((E2_CommandTaskData *)rt->extradata)->command, rt->pid,
					_("returned"), cmddata->exit);
				e2_output_print (rt->current_tab, message, rt->pidstr, TRUE, "small",
//					(WEXITSTATUS (rt->exit) == 0) ? "green" : "red", NULL);
					(cmddata->exit == 0) ? "green" : "red", NULL);
				g_free (message);
			}
			fd = g_io_channel_unix_get_fd (ioc);
			//shutdown the error input channel
			g_io_channel_shutdown (ioc, TRUE, NULL);
//crasher	g_io_channel_unref (ioc);
			e2_fs_close (fd);
			//shutdown the output channel
			fd = g_io_channel_unix_get_fd (cmddata->ioin);
			g_io_channel_shutdown (cmddata->ioin, TRUE, NULL);
			g_io_channel_unref (cmddata->ioin);
			e2_fs_close (fd);

			_e2_command_unblock_childsignal ();
		}
		return FALSE;	//removes the source
	}
	return TRUE;	//should never get to here
}
/**
 * @brief set up channel for receiving data from a child process
 * This is called twice for each async child (for child's stdout and stderr channels)
 * @param fd descriptor for the channel (from g_spawn_async_with_pipes())
 * @param cond flags setting the mode of channel operation
 * @param func the function to be called when @a cond is satisfied
 * @param rt pointer to data struct for the command to which the channel applies
 * @return the channel
 **/
static GIOChannel *_e2_command_setup_watch_channel
	(gint fd, GIOCondition cond, gpointer func, E2_TaskRuntime *rt)
{
	GIOChannel *ioc = g_io_channel_unix_new (fd);

	// Set IOChannel encoding to none, to make it fit for binary data
	//(which seems to mean that _any_ encoding will work)
	g_io_channel_set_encoding (ioc, NULL, NULL);
	g_io_channel_set_buffered (ioc, FALSE);
	g_io_channel_set_close_on_unref (ioc, TRUE);
	g_io_add_watch_full (ioc, e2_option_int_get ("command-watch-priority"),
		cond, (GIOFunc) func, rt, NULL);

	//undo reference added by g_io_add_watch ()
	g_io_channel_unref (ioc);
	return ioc;
}
/**
 * @brief find the pid of the last-started child command that's still running
 * @param anytab TRUE to find a match in any tab, FALSE in the currently-displayed tab
 * @return the PID or 0 if not found
 **/
guint e2_command_find_last_child (gboolean anytab)
{
	E2_TaskRuntime *rt;
	GList *member;
	pthread_mutex_lock (&task_mutex);
	member = g_list_last (app.taskhistory);
	pthread_mutex_unlock (&task_mutex);
	//backwards scan, don't bother locking at end of loop
	for (; member != NULL; member = member->prev)
	{
		rt = (E2_TaskRuntime *)member->data;
		if (rt != NULL && rt->status == E2_TASK_RUNNING && !rt->action
			&& (anytab || rt->current_tab == &app.tab))
			return rt->pid;
	}
	return 0;
}
/**
 * @brief kill child process with id @a pid
 * @param pid the id of the process we're looking for
 * @return TRUE if a child process with that pid was found and killed
 **/
gboolean e2_command_kill_child (guint pid)
{
	E2_TaskRuntime *rt = e2_task_find_task ((glong)pid);
	if (rt != NULL)
	{	//it's one of ours
		if (!kill ((pid_t) pid, SIGTERM))	//or SIGKILL
		{
			rt->status = E2_TASK_ABORTED;
			return TRUE;
		}
		e2_output_print_strerrno ();
	}
	return FALSE;
}
/**
 * @brief check whether process with id @a pid is running
 * @param pid the id of the process we're looking for
 * @return TRUE if a process with that pid is running
 **/
gboolean e2_command_find_process (guint pid)
{
	pid_t result = waitpid (pid, NULL, WNOHANG);
	return (result == 0);	//will be -1 if finished before, result if finished now
}
/**
 * @brief update all relevant child foreground-tab pointers after output pane tab change
 * @param currenttab
 * @param replacetab
 * @return
 **/
void e2_command_retab_children (E2_OutputTabRuntime *currenttab,
	E2_OutputTabRuntime *replacetab)
{
	E2_TaskRuntime *rt;
	GList *member;
	_e2_command_block_childsignal ();

	pthread_mutex_lock (&task_mutex);
	member = app.taskhistory;
	pthread_mutex_unlock (&task_mutex);
	while (member != NULL)
	{
		rt = (E2_TaskRuntime *) member->data;
		if (rt != NULL
			&& (rt->status == E2_TASK_RUNNING || rt->status == E2_TASK_PAUSED))
		{
			if (rt->current_tab == currenttab)
				rt->current_tab = rt->background_tab;
			else if (rt->background_tab == replacetab)
				rt->current_tab = currenttab;
		}
		pthread_mutex_lock (&task_mutex);
		member = member->next;
		pthread_mutex_unlock (&task_mutex);
	}

	_e2_command_unblock_childsignal ();
}
/**
 * @brief update tab ptrs of children using a deleted tab
 * @param currenttab the deleted tab's data struct
 * @param replacetab the replacement tab's data struct
 * @return
 **/
void e2_command_retab2_children (E2_OutputTabRuntime *currenttab,
	E2_OutputTabRuntime *replacetab)
{
	E2_TaskRuntime *rt;
	GList *member;
	_e2_command_block_childsignal ();

	pthread_mutex_lock (&task_mutex);
	member = app.taskhistory;
	pthread_mutex_unlock (&task_mutex);
	while (member != NULL)
	{
		rt = (E2_TaskRuntime *) member->data;
		if (rt != NULL
			&& (rt->status == E2_TASK_RUNNING || rt->status == E2_TASK_PAUSED)
			&& rt->background_tab == currenttab)
		{
			rt->background_tab = replacetab;
			rt->current_tab = replacetab;
		}
		pthread_mutex_lock (&task_mutex);
		member = member->next;
		pthread_mutex_unlock (&task_mutex);
	}

	_e2_command_unblock_childsignal ();
}
/**
 * @brief clear pending commands from queue
 * This is an action, results shown in the current output tab
 * @param from UNUSED the activated widget
 * @param art UNUSED runtime data sent to this action
 * @return TRUE if anything pending is removed
 **/
static gboolean _e2_command_clear_pending (gpointer from, E2_ActionRuntime *art)
{
	E2_TaskRuntime *rt;
	GList *member;
	gboolean retval = FALSE;
	_e2_command_block_childsignal ();
	//to avoid iteration issues, a 2-stage process ...
	pthread_mutex_lock (&task_mutex);
	member = app.taskhistory;
	pthread_mutex_unlock (&task_mutex);
	while (member != NULL)
	{
		rt = (E2_TaskRuntime *) member->data;
		if (rt->status == E2_TASK_QUEUED || rt->status == E2_TASK_NONE)
		{
			g_free (rt->pidstr);
			if (rt->action)
			{
				g_free (((E2_ActionTaskData*)rt->extradata)->currdir);
				g_free (((E2_ActionTaskData*)rt->extradata)->othrdir);
				GPtrArray *names = ((E2_ActionTaskData*)rt->extradata)->names;
				if (names != NULL)
					e2_fileview_clean_selected (names);
				gchar *data = ((E2_ActionTaskData*)rt->extradata)->rt_data;
				if (data != NULL)
					g_free (data);
			}
			else
			{
				g_free (((E2_CommandTaskData*)rt->extradata)->command);
				g_free (((E2_CommandTaskData*)rt->extradata)->currdir);
			}
			g_free (rt->extradata);
			g_free (rt);
			member->data = NULL;
			retval = TRUE;
		}
		pthread_mutex_lock (&task_mutex);
		member = member->next;
		pthread_mutex_unlock (&task_mutex);
	}
	pthread_mutex_lock (&task_mutex);
	app.taskhistory = g_list_remove_all (app.taskhistory, NULL);
	pthread_mutex_unlock (&task_mutex);
	_e2_command_unblock_childsignal ();
	return retval;
}
/**
 * @brief list queued pending commands
 * This is an action, results shown in the current output tab
 * @param from UNUSED the activated widget
 * @param art UNUSED runtime data sent to this action
 * @return TRUE if there is any pending command
 **/
static gboolean _e2_command_list_pending (gpointer from, E2_ActionRuntime *art)
{
//	printd (DEBUG, "e2_command_list_pending (data:)");
	//find start of waiting item(s) in Q
	E2_TaskRuntime *rt;
	GList *member;
	//quick check for something relevant
	pthread_mutex_lock (&task_mutex);
	member = app.taskhistory;
	pthread_mutex_unlock (&task_mutex);
	while (member != NULL)
	{
		rt = (E2_TaskRuntime *) member->data;
		if (rt->status <= E2_TASK_QUEUED)
			break;
		pthread_mutex_lock (&task_mutex);
		member = member->next;
		pthread_mutex_unlock (&task_mutex);
	}
	if (member == NULL)
	{
		e2_output_print (&app.tab, _("nothing is waiting"), NULL, TRUE, NULL);
		return FALSE;
	}
#ifdef E2_ASYNC
	gchar *srcdir;
#endif
	gchar *shortdir, *message;
	message = _("command");
	e2_output_print (&app.tab, message, NULL, TRUE, "bold", "uline", NULL);

	_e2_command_block_childsignal ();

	while (member != NULL)
	{
		rt = (E2_TaskRuntime *) member->data;
		if (rt != NULL && rt->status <= E2_TASK_QUEUED)
		{
#ifdef E2_ASYNC
			if (rt->action)
			{
				E2_ActionTaskData *atask = (E2_ActionTaskData *)rt->extradata;
				srcdir = F_FILENAME_FROM_LOCALE (atask->currdir);
				shortdir = e2_utils_str_shorten (srcdir, CMDNAMELEN, E2_DOTS_START);
				gchar *subject;
				if (g_str_has_prefix (atask->action->name, _A(5)))
				{	//this was a "file-action"
					subject =
					(atask->rt_data == NULL || *((gchar *)atask->rt_data) == '\0') ?
						_("<selected items>") : (gchar *)atask->rt_data;
				}
				else
					subject = "";	//this results in a redundant space in the line
				message = g_strdup_printf ("%s %s || %s",
					atask->action->name, subject, shortdir);
				F_FREE (srcdir);
			}
			else
#else
			if (!rt->action)
#endif
			{
#ifdef E2_ASYNC
				E2_CommandTaskData *ctask = (E2_CommandTaskData *)rt->extradata;
#endif
				shortdir =
				e2_utils_str_shorten (ctask->currdir, CMDNAMELEN, E2_DOTS_START);
				gchar *shortcmd =
				e2_utils_str_shorten (((E2_CommandTaskData *)rt->extradata)->command,
					CMDNAMELEN, E2_DOTS_END);
				message = g_strdup_printf (
					"%s || %s", shortcmd, shortdir);
				g_free (shortcmd);

			}
		}
		pthread_mutex_lock (&task_mutex);
		member = member->next;
		pthread_mutex_unlock (&task_mutex);
	}

	_e2_command_unblock_childsignal ();

	e2_output_print_end (&app.tab, FALSE);

	return TRUE;
}
/**
 * @brief count the active tasks of this session
 * This counts child processes and actions (queued or not) in the task history
 * list and marked as currently-running
 * @return the number of currently-running things
 **/
guint e2_command_count_running_tasks (gboolean countpaused)
{
	guint i = 0;
	E2_TaskRuntime *rt;
	GList *member;
	pthread_mutex_lock (&task_mutex);
	member = app.taskhistory;
	pthread_mutex_unlock (&task_mutex);
	while (member != NULL)
	{
		rt = (E2_TaskRuntime *) member->data;
		if (rt != NULL
			&& (rt->status == E2_TASK_RUNNING || (countpaused && (rt->status == E2_TASK_PAUSED))))
			i++;
		pthread_mutex_lock (&task_mutex);
		member = member->next;
		pthread_mutex_unlock (&task_mutex);
	}
	return i;
}
/**
 * @brief list the active child processes (commands and actions) of the current session
 * This is an action, results shown in the current tab
 * @param from UNUSED the activated widget
 * @param art UNUSED runtime data sent to this action
 * @return TRUE if there are any children
 **/
static gboolean _e2_command_list_children (gpointer from, E2_ActionRuntime *art)
{
//	printd (DEBUG, "_e2_command_list_children (data:)");
	if (e2_command_count_running_tasks (TRUE) == 0)
	{
		e2_output_print (&app.tab, _("nothing is running"), NULL, TRUE, NULL);
		return FALSE;
	}
#ifdef E2_ASYNC
	gchar *srcdir;
#endif
	gchar *shortdir, *message;
	message = _("  pid || directory || command");
	e2_output_print (&app.tab, message, NULL, TRUE, "bold", "uline", NULL);

	E2_TaskRuntime *rt;
#ifdef E2_ASYNC
	E2_ActionTaskData *task;
#endif
	GList *member;
	_e2_command_block_childsignal ();

	pthread_mutex_lock (&task_mutex);
	member = app.taskhistory;
	pthread_mutex_unlock (&task_mutex);
	while (member != NULL)
	{
		rt = (E2_TaskRuntime *) member->data;
		if (rt != NULL
			&& (rt->status == E2_TASK_RUNNING || rt->status == E2_TASK_PAUSED))
		{
#ifdef E2_ASYNC
			if (rt->action)
			{
				task = (E2_ActionTaskData *)rt->extradata;
				srcdir= F_FILENAME_FROM_LOCALE (task->currdir);
				shortdir = e2_utils_str_shorten (srcdir, CMDNAMELEN, E2_DOTS_START);
				message = g_strdup_printf (" %6lu || %s || %s %s", (gulong) rt->pid,
					shortdir, task->action->name,
					(task->rt_data == NULL) ?
						_("<selected items>") : (gchar *)task->rt_data);
				F_FREE (srcdir);
			}
			else
#else
			if (!rt->action)
#endif
			{
				shortdir = e2_utils_str_shorten (
				((E2_CommandTaskData *)rt->extradata)->currdir, CMDNAMELEN, E2_DOTS_START);
				message = g_strdup_printf (" %6ld || %s || %s", rt->pid, shortdir,
				((E2_CommandTaskData *)rt->extradata)->command);
			}

			e2_output_print (&app.tab, message, NULL, TRUE, NULL);
			g_free (shortdir);
			g_free (message);
		}
		pthread_mutex_lock (&task_mutex);
		member = member->next;
		pthread_mutex_unlock (&task_mutex);
	}

	_e2_command_unblock_childsignal ();

	e2_output_print_end (&app.tab, FALSE);

	return TRUE;
}
/**
 * @brief list the command history of the current session
 * This is an action, results shown in the current output tab
 * @param from UNUSED the activated widget
 * @param art UNUSED data sent to this action
 * @return TRUE if there is a history
 **/
static gboolean _e2_command_list_history (gpointer from, E2_ActionRuntime *art)
{
//	printd (DEBUG, "e2_command_list_history (data:)");
#ifdef E2_ASYNC
	gchar *srcdir;
#endif
	gchar *shortdir, *message;
	message = _("command || directory || result");
	e2_output_print (&app.tab, message, NULL, TRUE, "bold", "uline", NULL);

	E2_TaskRuntime *rt;
#ifdef E2_ASYNC
	E2_ActionTaskData *atask;
	E2_CommandTaskData *ctask;
#endif
	GList *member;
	_e2_command_block_childsignal ();

	pthread_mutex_lock (&task_mutex);
	member = app.taskhistory;
	pthread_mutex_unlock (&task_mutex);
	while (member != NULL)
	{
		rt = (E2_TaskRuntime *) member->data;
		if (rt != NULL && rt->status >= E2_TASK_COMPLETED)
		{
#ifdef E2_ASYNC
			if (rt->action)
			{
				atask = (E2_ActionTaskData *)rt->extradata;
				srcdir = F_FILENAME_FROM_LOCALE (atask->currdir);
				shortdir = e2_utils_str_shorten (srcdir, CMDNAMELEN, E2_DOTS_START);
				gchar *subject;
				if (g_str_has_prefix (atask->action->name, _A(5)))
				{	//this was a "file-action"
					subject =
					(atask->rt_data == NULL || *((gchar *)atask->rt_data) == '\0') ?
						_("<selected items>") : (gchar *)atask->rt_data;
				}
				else
					subject = "";	//this results in a redundant space in the line
				message = g_strdup_printf ("%s %s || %s || %s",
					atask->action->name, subject, shortdir,
					(atask->result) ? _("OK") : _("error"));
				F_FREE (srcdir);
			}
			else
#else
			if (!rt->action)
#endif
			{
				ctask = (E2_CommandTaskData *)rt->extradata;
				shortdir =
				e2_utils_str_shorten ((gchar *)ctask->currdir, CMDNAMELEN, E2_DOTS_START);
				gchar *shortcmd =
				e2_utils_str_shorten (((E2_CommandTaskData *)rt->extradata)->command,
					CMDNAMELEN, E2_DOTS_END);
				if (ctask->exit == 0)
					message = g_strdup_printf (
					"%s || %s || %s", shortcmd, shortdir, _("OK"));
				else
					message = g_strdup_printf (
					"%s || %s || %s %d", shortcmd, shortdir, _("error"), ctask->exit);
				g_free (shortcmd);
			}

			e2_output_print (&app.tab, message, NULL, TRUE, NULL);
			g_free (shortdir);
			g_free (message);
		}
		pthread_mutex_lock (&task_mutex);
		member = member->next;
		pthread_mutex_unlock (&task_mutex);
	}

	_e2_command_unblock_childsignal ();

	e2_output_print_end (&app.tab, FALSE);

	return TRUE;
}
/**
 * @brief pipe utf-8 string @a message to process @a pid
 * The specified process must be a child of the current e2 instance
 * @param pid the id of the destination process, or 0 for the last-created child
 * @param message utf string to send to process @a pid
 * @return TRUE if the string was sent without error
 **/
static gboolean _e2_command_send_to_pid (guint pid, gchar *message)
{
	gchar *errmsg;
	printd (DEBUG, "e2_command_send_to_pid (pid:%d,messge:%s)", pid, message);
	E2_TaskRuntime *rt;
	if (pid == 0)
	{
		pid = e2_command_find_last_child (FALSE);
		if (pid == 0)
			return FALSE;	//FIXME warning
	}
	rt = e2_task_find_task (pid);
	if (rt == NULL || rt->status != E2_TASK_RUNNING || rt->action)
	{
		errmsg = g_strdup_printf (_("The process with pid %d is not our child"), pid);
		e2_output_print_error (errmsg, TRUE);
		g_free (errmsg);
		return FALSE;
	}

	gsize size;
	GIOStatus stat;
	GError *error = NULL;
	GIOChannel *channel = ((E2_CommandTaskData *)rt->extradata)->ioin;
	//CHECKME what if the process doesn't understand utf-8 ?
	stat = g_io_channel_write_chars (channel, message, -1, &size, &error);
	if (error != NULL)
	{
		errmsg = g_strdup_printf (_("Failed writing to child: %s"), error->message);
		printd (WARN, errmsg);
		e2_output_print (rt->current_tab,
			errmsg, rt->pidstr, TRUE, E2_ERRORTAGS, NULL);
		g_error_free (error);
		g_free (errmsg);
		return FALSE;
	}
	error = NULL;
	stat = g_io_channel_write_chars (channel, "\n", 1, &size, &error);
	if (error != NULL)
	{
		errmsg = g_strdup_printf (_("Failed writing to child: %s"), error->message);
		printd (WARN, errmsg);
		e2_output_print (rt->current_tab,
			errmsg, rt->pidstr, TRUE, E2_ERRORTAGS, NULL);
		g_error_free (error);
		g_free (errmsg);
		return FALSE;
	}
	else if (((E2_CommandTaskData *)rt->extradata)->show)
	{	//signal what has been sent
		errmsg = g_strconcat (message, "\n", NULL);
		e2_output_print (rt->current_tab,
			errmsg, rt->pidstr, FALSE, NULL);
		g_free (errmsg);
	}
	return TRUE;
}
/**
 * @brief check whether @a command has any "modifiers"
 * @param command utf command string
 * @param macros pointer to flag, set FALSE if leading '!' found
 * @param shell pointer to flag, set TRUE if leading '>' found
 * @param wait pointer to flag, set TRUE if leading '|' found
 * @param show pointer to flag, set FALSE if trailing '&' found
 * @return integer, the byte-offset into @a command where the next char to be processed is
 **/
static gint _e2_command_run_checks (gchar *command,
	gboolean *macros, gboolean *shell, gboolean *wait, gboolean *show)
{
	gint retval = 0;
	gboolean done = FALSE;
	//checks here are all for single-gchar characters, no need for utf functions
	while (!done)
	{
		switch (command[0])
		{
			case ' ':	//skip leading whitespace
			case '\t':
				break;
			case '!':
				*macros = FALSE;
				break;
			case '>':
				*shell = TRUE;
				break;
			case '|':
				*wait = TRUE;
				break;
			default:
				done = TRUE;
				break;
		}
		if (!done)
		{
			retval++;
			command++;
		}
	}
	if (!*shell)
	{	//check for redirection chars, force shell in that case
		gchar *s = e2_utils_bare_strchr (command, '>');
		if (s != NULL && s > command && *(s+1) != '\0')
			*shell = TRUE;
		else if ((s = e2_utils_bare_strchr (command, '|')) != NULL
			&& s > command && *(s+1) != '\0')
			*shell = TRUE;
		else if ((s = e2_utils_bare_strchr (command, '<')) != NULL
			&& s > command && *(s+1) != '\0')
			*shell = TRUE;
	}
	if (g_str_has_suffix (command, "&"))
	{
		*show = FALSE;
		command[strlen (command) - sizeof (gchar)] = '\0';
	}
	return retval;
}
/**
 * @brief asynchronously run external command
 *
 * @param command the command string, used in messages, utf8
 * @param args pointer to command arguments array, each localised
 * @param show TRUE to show @a command and its pid in the output pane
 * @param prefix string to add to the start of @a command for its display
 *
 * @return the pid of process running the command, or 0 if there's an error
 **/
#if DEBUG_LEVEL > 2
GIOChannel *stdoutchannel;
GIOChannel *stderrchannel;
#endif
static gint _e2_command_run_async (gchar *command, gchar **args,
	gboolean show, gchar *prefix)
{
	printd (DEBUG, "asynchronously run command %s", command);
	_e2_command_block_childsignal ();	//WHY block SIGCHLD signals here?

	gchar *local = F_FILENAME_TO_LOCALE (curr_view->dir);
	GPid pid = 0;
	gint pin = 0, pout = 0, perr = 0;
	GError *error = NULL;
	g_spawn_async_with_pipes (local, args, NULL,
		G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH,
		NULL, NULL, &pid, &pin, &pout, &perr, &error);
	F_FREE (local);

	if (error != NULL)
	{
		gchar *message = g_strconcat (prefix, command, " ", NULL);
		//always use the current tab at the start of the command
		e2_output_print (&app.tab, message, NULL, TRUE, "bold", "blue", NULL);
		g_free (message);
		message = g_strconcat (error->message, ".", NULL);
		e2_output_print_error (message, TRUE);
		g_error_free (error);
		_e2_command_unblock_childsignal ();
		return 0;
	}

	GIOChannel *ioc = g_io_channel_unix_new (pin);	//channel for sending to child stdin
	g_io_channel_set_encoding (ioc, NULL, NULL);	//binary (=any) encoding
	g_io_channel_set_buffered (ioc, FALSE);
	g_io_channel_set_close_on_unref (ioc, TRUE);
	//FIXME manage the removal of channels' event sources

	E2_TaskRuntime *rt = NULL;

	//keep trying this until data set fn not busy
	while ((rt = e2_task_set_data (pid, E2_TASKTYPE_ASYNC, command, 0, ioc)) == NULL);

	if (rt == GINT_TO_POINTER (1))
	{
		//FIXME cleanup
		_e2_command_unblock_childsignal ();
		return 0;
	}

	((E2_CommandTaskData *)rt->extradata)->show = show;

	if (show && e2_option_bool_get ("fileop-show"))
	{
		gchar *message = g_strconcat (prefix, command, NULL);
		e2_output_print (rt->current_tab,
			message, rt->pidstr, FALSE, "bold", "blue", NULL);
		g_free (message);
		message = g_strdup_printf ("  (%d)\n", pid);
		e2_output_print (rt->current_tab,
			message, rt->pidstr, FALSE, "small", "blue", NULL);
		g_free (message);
	}
	//channel for getting the child's stdout
#if DEBUG_LEVEL > 2
	stdoutchannel =
#endif
	_e2_command_setup_watch_channel (pout,
		G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
		_e2_command_watch_std, rt);
	//channel for getting the child's stderr
#if DEBUG_LEVEL > 2
	stderrchannel =
#endif
	_e2_command_setup_watch_channel (perr,
		G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
		_e2_command_watch_err, rt);

	_e2_command_unblock_childsignal ();

	return pid;
}
/**
 * @brief run command @a command synchronously (block until it's finished)
 *
 * @param command the command string, used in messages, utf8
 * @param args pointer to command arguments array, each localised
 * @param show TRUE to show @a command and its exit status in the output pane
 * @param prefix string to add to the start of @a command for its display
 *
 * @return the exit code of the command, > 0 if error
 **/
static gint _e2_command_run_sync (gchar *command, gchar **args,
	gboolean show, gchar *prefix)
{
	printd (DEBUG, "synchronously run command %s", command);
	//we dont get real pid's from static commands, so we fake
	//by using -ve ones in simple sequence
	static glong fakepid = 0;
	fakepid--;

	gchar *message;
	E2_TaskRuntime *rt = NULL;

	//keep trying this until data set fn not busy
	while ((rt = e2_task_set_data (fakepid, E2_TASKTYPE_SYNC, command, 0, NULL)) == NULL);

	if (rt == GINT_TO_POINTER (1))
	{
		//CHECKME warn user ?
		fakepid++;
		return 1;
	}
	//but mark-names starting with "-" are confusing so substitute "s"
	*(rt->pidstr) = 's';

	if (show && e2_option_bool_get ("fileop-show"))
	{
		gchar *message = g_strconcat (prefix, command, NULL);
		e2_output_print (rt->current_tab,
			message, rt->pidstr, FALSE, "bold", "blue", NULL);
		g_free (message);
		message = g_strdup_printf ("  (%s)\n", rt->pidstr);
		e2_output_print (rt->current_tab,
			message, rt->pidstr, FALSE, "small", "blue", NULL);
		g_free (message);
		WAIT_FOR_EVENTS;
	}

//	GError *error;
	gint exit;
	gchar *sout = NULL, *serr = NULL;
	gchar *local = F_FILENAME_TO_LOCALE (curr_view->dir);

	_e2_command_block_childsignal ();

	rt->status = E2_TASK_RUNNING;
	gboolean success = g_spawn_sync (local, args, NULL,
		G_SPAWN_SEARCH_PATH, NULL, NULL, &sout, &serr, &exit, NULL);	//&error);
	rt->status = E2_TASK_COMPLETED;

	_e2_command_unblock_childsignal ();

	((E2_CommandTaskData *)rt->extradata)->exit = exit;

	if (success && show && sout != NULL && *sout != '\0')
	{
		e2_output_print (rt->current_tab, sout, rt->pidstr, FALSE, NULL);
	}
	if (success && serr != NULL && *serr != '\0')
	{
		e2_output_print_error (serr, FALSE);
	}
	if (sout != NULL)
		g_free (sout);
	if (serr != NULL)
		g_free (serr);
	F_FREE (local);

	if (success && show && e2_option_bool_get ("fileop-show"))
	{
		message = g_strdup_printf ("%s%s (%s) %s '%d'\n", prefix, command,
			rt->pidstr, _("returned"), WEXITSTATUS (exit));
		e2_output_print (rt->current_tab, message, rt->pidstr, TRUE, "small",
			(WEXITSTATUS (exit) == 0) ? "green" : "red", NULL);
		g_free (message);
	}
	if (!success)
	{
		message = g_strdup_printf (_("Strange error: could not run %s"), command);
		e2_output_print_error (message, TRUE);
	}

//	_e2_command_unblock_childsignal ();
	return WEXITSTATUS (exit);
}
/**
 * @brief run a single command @a raw, after interpreting its contents
 * Downstream functions expect BGL to be closed
 * Command modifiers (!>&) are interpreted.
 * $... variables are interpreted
 * %... macros are expanded unless the "no-expand" modifier '!' is present at
 * the start of @a raw.
 * Any 'pid:' directive is interpreted. The 'pid' string must describe a base 10 integer.
 * Any "\" that doesn't follows another "\" is simply removed.
 * Any "~" followed by "/" is interpreted as $HOME, as is a "~" by itself.
 * If * or ? is in any argument, that will be expanded, if possible, into a
 * series of matching itemnames in the active directory.
 * "cd" commands that are not run in a separate shell just change the
 * dir shown in the active pane. In that case the new dir must be
 * separated by one or more ' ' chars (ie no tabs), or "cd" alone is
 * interpreted as "cd $HOME".
 * Command arguments for must be separated by one or more ' ' chars (ie no tabs)
 * NOTE used to return pid of process running the command (if any)
 * @param raw utf command string
 * @param range flags for the scope of the command
 *
 * @return 0 for success, 1 for error
 **/
//FIXME more consistent returned value
static gint _e2_command_run_single (gchar *raw, E2_CommandRange range)
{
	//CHECKME should variables be volatilized for threaded usage ?
	gchar *command1, *command2, *freeme;
	gboolean expand_macros = TRUE;
	gboolean run_in_shell = FALSE;
	gboolean wait = FALSE;
	gboolean show = TRUE;
	gint pid;

	//1st, find how the command is to be run
	command1 = raw + _e2_command_run_checks
		(raw, &expand_macros, &run_in_shell, &wait, &show);
	//(wait and show will be overridden, later, if range == E2_COMMAND_RANGE_FILE_ACTION)

	//replace any alias if so desired
	if (//CHECKME range != E2_COMMAND_RANGE_FILE_ACTION //this is not an action on selected items
		//&& expand_macros	//no instruction to not expand macros
		//&&
		e2_option_bool_get ("command-use-aliases")	//alias use is expected
	)
	{
		freeme = e2_alias_apply (command1, &app.aliases);
		//check command operation mode again, in case alias changed it
		command1 = freeme + _e2_command_run_checks (freeme,
			&expand_macros, &run_in_shell, &wait, &show);
	}
	else
		freeme = NULL;

	E2_Action *action = e2_action_check (command1);
	gboolean command_is_action = (action != NULL);
	if (command_is_action)
	{
		if (action->exclude & E2_ACTION_EXCLUDE_ACCEL)
		{
			gchar *msg = g_strdup_printf (_("Cannot run '%s'"), command1);
			e2_output_print_error (msg, TRUE);
			if (freeme != NULL)
				g_free (freeme);
			return 1;
		}
		//set the correct action-range according to whether the
		//name begins with "file" and has no argument
		if (g_str_has_prefix (command1, _A(5)))
		{
			command2 = e2_utils_find_whitespace (command1);
			if (command2 == NULL || e2_utils_pass_whitespace (command2) == NULL)
				range = E2_COMMAND_RANGE_FILE_ACTION;
		}
		else if (range == E2_COMMAND_RANGE_FILE_ACTION)
			range = E2_COMMAND_RANGE_DEFAULT;
	}

	//set some default operating modes for selected-items commands
	//(overrides show flag in raw or alias, if any)
	if (!command_is_action && range == E2_COMMAND_RANGE_FILE_ACTION)
	{
	//CHECKME is this of any real use ? it's intended mainly for actions
	//	wait = !e2_option_bool_get ("fileop-background");
		show = e2_option_bool_get ("fileop-show");
	}

	//if appropriate, arrange to append the selected item(s),
	//as the (or additional) command argument(s)
	if (range == E2_COMMAND_RANGE_FILE_ACTION	//this is an action on selected items
		&& expand_macros	//no instruction to not expand macros
		&& g_str_has_prefix (command1, _A(5))	//this really is a file action
/*		//FIXME support macros other than with english letters
		&& strstr (command1, "%f") == NULL
		&& strstr (command1, "%F") == NULL
		&& strstr (command1, "%p") == NULL
		&& strstr (command1, "%P") == NULL
*/	)
		//if there is no filelist macro already in the command, add one
//		command2 = g_strconcat (command1, " %f", NULL);
	{
		//if there is no command argument, append " %f"
		command2 = e2_utils_find_whitespace (command1);
		if (command2 != NULL)
			command2 = e2_utils_pass_whitespace (command2);
		if (command2 == NULL)
			command2 = g_strconcat (command1, " %f", NULL);
		else
			command2 = g_strdup (command1);
	}
	else
		command2 = g_strdup (command1);
	if (freeme != NULL)
		g_free (freeme);
	//newly-allocated command2 is now the focus ..

	//replace variables etc
	command1 = e2_utils_replace_vars (command2);
	g_free (command2);
	//replace wildcards
	if (!run_in_shell && expand_macros)
		command1 = e2_utils_replace_wildcards (command1);

/*	//as a convenience, eliminate escape chars from the command
	//(emelfm2 commands per se do not need anything escaped_
	if (strchr (command1, '\\') != NULL)
	{
		command2 = command1;
		while ((command2 = strchr (command2, '\\')) != NULL)
		{
			if (*(command2+1) != '\\')
				*command2 = '\1';
			command2++;
		}
		if (strchr (command1, '\1') != NULL)
		{
			gchar old[2] = {'\1', '\0'};
			command2 = command1;
			command1 = e2_utils_str_replace (command1, old, "");
			g_free (command2);
		}
	} */
	//now newly-allocated command1 has focus, command2 is clean

	//if command is: cd [%]%f, when updir '../' line was clicked, that
	//item would not be available as a selected item, so ...
	if (g_str_has_prefix (command1, "cd"))
	{
		//FIXME support macros other than with english letters
		if (strstr (command1, "%f") != NULL)
		{
			FileInfo *info = e2_fileview_get_selected_first_local (curr_view);
			if (info != NULL)
			{
				if (g_str_equal (info->filename, ".."))	//ascii, no need for utf
				{
					g_free (command1);
					command1 = g_strdup ("cd ..");	//no translation
					expand_macros = FALSE;
				}
			}
		}
	}
	//expand macros, if wanted
	if (expand_macros)
	{
		GString *str = e2_utils_expand_macros (command1, NULL);
		if (str == NULL)
		{
			e2_output_print_error (_("Failed to expand macros"), FALSE); //current tab
			pid = 1;
			goto cleanup;
		}
		else if (str == GINT_TO_POINTER(1))
		{	//the user cancelled in a prompt macro
			pid = 1;
			goto cleanup;
		}
		g_free (command1);
		command1 = str->str;
		g_string_free (str, FALSE);
	}

	//interpret any 'pid:' directive
	//the 'pid' string must describe a base 10 integer, no
	//bigger than the max integer for the current system

	//i.e. should be 5 bytes long (16385) for 32-bit systems,
	//10 bytes for 64-bit systems
//#if sizeof(gint) == 4
//#define E2_PID_SIZE 5
//#else
//#define E2_PID_SIZE 10
//#endif

	command2 = NULL;
	if (tochild)
	{
		pid = (_e2_command_send_to_pid (0, command1)) ? 0 : 1;
		goto cleanup;
	}
	if (//range != E2_COMMAND_RANGE_FILE_ACTION	//CHECKME why excluded ?
		//&&
		(command2 = strchr (command1, ':')) != NULL	//if always ascii ':', don't need g_utf8_strchr()
	)
	{
		if (command2 > command1)	// there was something ahead of the ':'
		{
			*command2 = '\0';
			gchar *end;
			gint i = (gint) g_ascii_strtoull (command1, &end, 10);
			if (end == command2)
				//the whole string was used in conversion, so it was
				//a valid number and nothing else
			{
				printd (DEBUG, "sending '%s' to process '%s'", command2+1, command1);
				pid = (_e2_command_send_to_pid (i, command2+1)) ? 0 : 1;	//+1 to skip the '\0'
				goto cleanup;
			}
			*command2 = ':';
		}
	}

	//"cd" commands that are not run in a separate shell
	//just change the dir in the active pane
	if (!run_in_shell && g_str_has_prefix (command1, "cd"))	//no translation
	{	//note: tests here rely on ascii single-byte chars in command string
		gint len = strlen (command1);
		if ((len > 3) && (command1[2] == ' '))	//FIXME or \t ?
		{
			e2_pane_change_dir (NULL,
#ifdef E2_OKVFSTMP
				FS_SAME, NULL,
#endif
				command1+3);	//CHECKME command1 utf-8 or ascii
			pid = 0;
			goto cleanup;
		}
		if (len == 2)
		{
/*			gchar *home = (gchar *) g_get_home_dir ();
			e2_pane_change_dir (NULL,
#ifdef E2_OKVFSTMP
				FS_OPENLOCAL, NULL,
#endif
				home);
*/
			e2_pane_change_dir (NULL,
#ifdef E2_OKVFSTMP
				FS_OPENLOCAL, NULL,
#endif
				"~");
			pid = 0;
			goto cleanup;
		}
	}
	else if /*(g_str_has_prefix (command1, "<"))
	{	//action
		if (strchr (command1, '.') == NULL)	//quick check
		{
			e2_output_print_error (_("You must specify an action to run, like '<file.copy'"), FALSE); //current tab
			pid = 1;
			goto cleanup;
		}
		if (strchr (command1, '.') != NULL)	//quick check */
		(command_is_action)
		{
			gchar *gap = e2_utils_find_whitespace (command1);
			if (gap != NULL)
			{
				*gap = '\0';
				gap = e2_utils_pass_whitespace (gap+1);
			}
			pid = (e2_action_run_simple (command1, gap)) ?	//now gap points to args, or NULL
				0 : 1;
			goto cleanup;
		}
//	}

//	gboolean free;
	gchar **args;
	gint argc;
	if (run_in_shell)
	{	//run in new shell
//		free = TRUE;	//free args, and separately free command1
		argc = 3;
		//construct freeable vector of args
//		gchar *args_ = g_strconcat (shellcmd,"\1-c\1", command1, NULL);
//		args = g_strsplit (args_, "\1", argc);
		args = g_malloc (sizeof (gchar *) * 4);
		args[0] = g_strdup (shellcmd);
		args[1] = g_strdup ("-c");
		args[2] = g_strdup (command1);
		args[3] = NULL;
	}
	else
	{	//run in same shell */
		//g_shell_parse_argv() leaves one of the args = command1
		//so don't trigger a double-free
	//	free = FALSE;	//free args, but not doubly-so for command1
		GError *error = NULL;
		//NB the glib 2.6 API doco is wrong about how this works !!
		g_shell_parse_argv (command1, &argc, &args, &error);
		if (error != NULL)
		{
			gchar *error_message = g_strdup_printf ("%s '%s' - ", _("Failed parsing command line"), command1);
			e2_output_print (&app.tab, error_message, NULL, FALSE, E2_ERRORTAGS, NULL);
			e2_output_print_strerrno ();
			g_free (error_message);
			g_error_free (error);
			pid = 1;
			goto cleanup;
		}
	}

#ifndef FILES_UTF8ONLY
	if (!app.utf8_filenames)
	{
//		free = TRUE;	//free all args anyway
		//localise the actual command strings
		gchar *local;
		gchar **tmp = args;
		while (*tmp != NULL)
		{	//we always take a copy and strfreev later
			local = D_FILENAME_TO_LOCALE (*tmp);
			g_free (*tmp);
			*tmp = local;
			tmp++;
		}
	}
#endif
	gchar *prefix = run_in_shell ? "sh>" : ">";
	pid = (wait) ?
		_e2_command_run_sync (command1, args, show, prefix) :
		_e2_command_run_async (command1, args, show, prefix);

//	if (free)
		g_free (command1);
	g_strfreev (args);

	return ((pid == 0) ? 1 : 0);

cleanup:
	g_free (command1);
	return pid;
}

/**
 * @brief send command to last child process, if any
 * Various checks are performed, to decide whether it's ok to send the command
 *
 * @param action_data UNUSED ptr to data assigned when action struct created at session start
 * @param rt_data UNUSED ptr to data sent to this command
 * @param widget the entry widget that was activated to trigger the action
 *
 * @return TRUE if task completed successfully, else FALSE
 */
static gboolean _e2_command_send_to_lastchild (gpointer from, E2_ActionRuntime *art)
{
	gchar *msg;
	//get the command string
	const gchar *command_raw = gtk_entry_get_text (GTK_ENTRY (from));
	//quick exit
	if (command_raw == NULL || *command_raw == '\0')
		return FALSE;
//	if (strchr (command_raw, ';') != NULL)
//		goto error;
	if (e2_command_find_last_child (FALSE) == 0)
		goto error;
	gchar *command = g_strdup (command_raw);
	if (e2_action_check (command) != NULL)
	{
		g_free (command);
		goto error;
	}
	tochild = TRUE;
	//this is an action, can only return T/F, so we lose the result of the command
	_e2_command_run_single (command, E2_COMMAND_RANGE_DEFAULT);
	tochild = FALSE;
	//we don't want any history, just clear the text
	gtk_entry_set_text (GTK_ENTRY (from), "");
	g_free (command);
	return TRUE;

error:
	msg =
	g_strdup_printf (_("Cannot send \"%s\" to a child process"), command_raw);
	e2_output_print_error (msg, TRUE);
	return FALSE;
}

  /****************/
 /**** public ****/
/****************/

/**
 * @brief run command @a command, after interpreting its contents
 *
 * @a command may be a single command or a series joined by ';'
 * (any ';' inside single- or double-parentheses is ignored)
 * NOTE used to return command's pid
 * @param command utf string containing the command(s)
 * @param range flags for the scope of the command (may be over-ridden, for actions)
 * @return integer return code from last command executed, 0 or 1
 **/
gint e2_command_run (gchar *command, E2_CommandRange range)
{
	//CHECKME should these variables be volatilized for threaded usage ?
	if ((command == NULL) || (*command == '\0'))
		return 1;
	g_strstrip (command);
	if (*command == '\0')
		return 1;

	printd (DEBUG, "e2_command_run (command:%s,range:%d)", command, range);
	//parse & handle joined commands here
	//in particular, ignore separators inside parentheses
	if (strchr (command, ';') == NULL)	//if always ascii ;, don't need g_utf8_strchr()
		//no joined commands
		return _e2_command_run_single (command, range);
	gint result = 0;
	gchar *s = command;
	gchar *p;
	while ((p = e2_utils_bare_strchr (s, ';')) != NULL)
	{
		//process this command
		*p = '\0';
		result = _e2_command_run_single (s, range);
		s = p; s++;	//resume from the next spot
		if (*s == '\0') //the ';' was at end of line
			return result;
	}
	//process the last (or only) command
	return (_e2_command_run_single (s, range));
}
/**
 * @brief display commands help document
 * The document name is hard-coded with the help-file name
 * defined at build-time
 * @return
 **/
void e2_command_output_help (void)
{	//check whether there's still an alias for the help command
	//(and BTW get a copy of the string, which needs to be freeable for e2_command_run)
	gchar *help_cmd = _("help");
	gchar *command = e2_alias_apply (help_cmd, &app.aliases);
	if (g_str_equal (help_cmd, command))
	{	//we can't find what the user has set for a help command
		//so use the default instead
		g_free (command);
		command = g_strconcat(_A(5),".",_A(92),
//			" "DOC_DIR, MAIN_HELP" [",_("commands"),"]",NULL);	//file.view_at
			" ", e2_option_str_get ("usage-help-doc"), " [",_("commands"),"]",NULL);	//file.view_at
	}
	e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
	g_free (command);
}
/**
 * @brief register command-related actions
 * @return
 **/
void e2_command_actions_register (void)
{
	//not action stuff, but this is a convenient spot for other init
	_e2_command_set_sigchild_handler ();
	struct passwd *pw;
	pw = getpwuid (getuid ());
	if (pw != NULL)
	{
		gchar *baseshell = g_path_get_basename (pw->pw_shell);
		shellcmd = g_find_program_in_path (baseshell);
		g_free (baseshell);
		if (shellcmd == NULL)
			shellcmd = "/bin/sh";
	}
	else
		shellcmd = "/bin/sh";

	gchar *action_name = g_strconcat(_A(28),".",_A(22),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_list_children, NULL, FALSE);
	action_name = g_strconcat(_A(50),".",_A(22),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_list_history, NULL, FALSE);
	action_name = g_strconcat(_A(65),".",_A(22),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_list_pending, NULL, FALSE);
	action_name = g_strconcat(_A(65),".",_A(38),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_clear_pending, NULL, FALSE);
	action_name = g_strconcat(_A(1),".",_A(76),NULL);
	e2_action_register (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_send_to_lastchild, NULL, FALSE,
		E2_ACTION_EXCLUDE_TOOLBAR | E2_ACTION_EXCLUDE_MENU, NULL);
}
/**
 * @brief register command-related options
 * @return
 **/
void e2_command_options_register (void)
{
	//no screen rebuilds needed after any change to these options
	gchar *group_name = g_strconcat(_C(5),":",_C(24),NULL); //_("commands:misc"
	e2_option_str_register ("command-xterm", group_name, _("x terminal emulator:"),
		_("This is the external command/application that will be be run when emelFM2 is asked to open a terminal"),
		NULL, "xterm",
		E2_OPTION_FLAG_BASIC | E2_OPTION_FLAG_FREEGROUP);
	e2_option_bool_register ("use-external-viewer", group_name, _("use external file-viewer"),
		_("If activated, the command entered below will be run to view file content, instead of launching the internal viewer"),
		NULL, FALSE,
		E2_OPTION_FLAG_BASIC);
	e2_option_str_register ("command-viewer", group_name, _("viewer command:"),
		_("This is a command to run an external application for viewing file content.\n"
		"The first selected item will be supplied as the first argument"),    //CHECKME = this is wrong ??
		"use-external-viewer", "gview",
		E2_OPTION_FLAG_BASIC);
	e2_option_bool_register ("use-external-editor", group_name, _("use external file-editor"),
		_("If activated, the command entered below will be run to edit file content, instead of launching the internal editor"),
		NULL, FALSE,
		E2_OPTION_FLAG_BASIC);
	e2_option_str_register ("command-editor", group_name, _("editor command:"),
		_("This is a command to run an external application for editing file content.\n"
		"The first selected item will be supplied as the first argument"),    //CHECKME = this is wrong ??
		"use-external-editor", "leafpad",
		E2_OPTION_FLAG_BASIC);
	e2_option_bool_register ("use-external-encoder", group_name, _("use external encoding converter"),
		_("If activated, the command entered below will be run, instead of using the internal conversion functions, to convert file character encoding when needed"),
		NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED);
	gchar *defcmd = g_strconcat ("enca -L %{", _("File encoding"),":} -x UTF8 <%f", NULL);
	e2_option_str_register ("command-encoder", group_name, _("converter command:"),
		_("A command which runs an external application to convert text encoding to UTF-8"),
		"use-external-encoder", defcmd,
		E2_OPTION_FLAG_ADVANCED);
	g_free (defcmd);
	e2_option_bool_register ("command-use-aliases", group_name, _("use aliases"),
		_("This is a general switch to turn on/off alias handling for commands"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("transparent-dir-links", group_name, _("interpret 'relative' paths"),
		_("This enables correct interpretation of paths containing '..' etc. These might be typed in, or attached to a button, say"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_int_register ("command-watch-priority", group_name, _("watch priority"),
		_("The watch priority of commands influences how fast program\n"
		"output is read from i/o channels. A too-high priority might\n"
		"decrease gui responsiveness and break automatic scrolling.\n"
		"(note: negative values mean high priority, positive values mean low priority)"),
		NULL, 300, -200, 400,
		E2_OPTION_FLAG_ADVANCED);

	//these options relate to actions, not commands
	group_name = g_strconcat(_C(5),":",_C(12),NULL);  //_("commands:file operations"
#ifdef E2_ASYNC
	e2_option_bool_register ("task-timeout-checks", group_name, _("stop after timeout"),
	_("If activated, each file operation will be terminated if not finished within the time-interval set below"),
	NULL, FALSE,
	E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);
	e2_option_int_register ("task-timeout-interval", group_name, _("timeout interval"),
	_("The interval (seconds) allowed to complete any file operation"),
	"task-timeout-checks", 20, 1, 3600,
	E2_OPTION_FLAG_ADVANCED);
#else
	e2_option_bool_register ("fileop-background", group_name, _("run in background"),
		_("This causes commands operating on the selected items to run in the background, not blocking the interface"),
		NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);
#endif
	e2_option_bool_register ("confirm-delete", group_name, _("confirm any delete"),
		_("If activated, you will be asked for confirmation before actually deleting anything"), NULL, TRUE,
		E2_OPTION_FLAG_BASIC);
	e2_option_bool_register ("confirm-overwrite", group_name, _("confirm any overwrite"),
		_("If activated, you will be asked for confirmation before actually overwriting anything"), NULL, TRUE,
		E2_OPTION_FLAG_BASIC);
	e2_option_bool_register ("relative-symlinks", group_name, _("relative symlinks"),
		_("This gives each created symlink a relative path to its source, like '../../<otherbranch>/<file>', instead of a full path referenced to /"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
}
