/* $Id: e2_file_info_dialog.c 2190 2006-07-31 04:31:54Z tpgww $
 *
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 * Portions copyright (C) 1999 Michael Clark.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "emelfm2.h"
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include "e2_dialog.h"

//max length of the file type string
#define FILE_TYPE_LENGTH 128

//filename is localised
static void _e2_file_info_dialog_get_file_type (gchar *filename, gchar *string,
	gint  len)
{
	gchar *command = g_strdup_printf ("file \"%s\"", filename);
	E2_FILE *pipe = e2_fs_open_pipe (command);
	g_free (command);
	if (pipe == NULL)
		return;

	size_t bsize = len+NAME_MAX;
	gchar *buf = (gchar *) g_malloc ((gulong) bsize);
#ifdef __USE_GNU
	ssize_t bytes_read = getdelim (&buf, &bsize, '\n', pipe);
	if (bytes_read >= 2)
#else //getdelim() not available
	gchar *strtmp = fgets (buf, bsize, pipe);
	if (strtmp != NULL && *strtmp != '\n')
#endif
	{
		gchar *s;
		if ((s = strchr (buf, ':')) != NULL)	//if always ascii ':', don't need g_utf8_strchr()
			g_strlcpy (string, s+2, len);
		else
			g_strlcpy (string, buf, len);
	}
	e2_fs_pipe_close (pipe);
	g_free (buf);
}

/**
 * @brief dialog response callback
 *
 * @param dialog the permissions-dialog widget
 * @param response the response for the clicked button
 * @param result pointer to store for the dialog return value
 *
 * @return
 **/
static void _e2_file_info_dialog_response_cb (GtkWidget *dialog,
	gint response, DialogButtons *result)
{
	if (response == GTK_RESPONSE_CLOSE)	//not covered in the general decoder
		*result = CANCEL;
	else
		e2_dialog_response_decode_cb (NULL, response, result);
	gtk_widget_destroy (dialog);
	gtk_main_quit ();
}

DialogButtons e2_file_info_dialog_run (gchar *localpath, gboolean multi)
{
	GtkWidget *dialog;
	GtkWidget *dialog_vbox;
	GtkWidget *table;
	GString *label_text = g_string_sized_new (NAME_MAX + 32);  //this is plenty big for all uses here
	gchar perm_string[32];
	gchar date_string[32];
    struct stat statbuf;
	struct tm *tm_ptr;
	struct passwd *pw_buf;
	struct group *grp_buf;

	e2_fs_lstat (localpath, &statbuf);
	//must prevent the default dialog response, it stuffs up Q cleanups
	DialogButtons choice;
	dialog = e2_dialog_create (NULL, NULL, _("file info"),
		_e2_file_info_dialog_response_cb, &choice);
	dialog_vbox = GTK_DIALOG (dialog)->vbox;
	gtk_container_set_border_width (GTK_CONTAINER (dialog_vbox), E2_PADDING);
	gchar *name = g_filename_display_basename (localpath);
	g_string_printf (label_text, "%s: <b>%s</b>", _("Name"), name);
	e2_widget_add_label (dialog_vbox, label_text->str, 0, TRUE, E2_PADDING);
	e2_widget_add_separator (dialog_vbox, TRUE, E2_PADDING);

	gchar *s1 = _("Type");
	gchar *s2, *tmsg;

	if (S_ISLNK (statbuf.st_mode))
		s2 = _("Symbolic Link");
	else if (S_ISREG (statbuf.st_mode))
	{
		gchar file_type[FILE_TYPE_LENGTH];
		gint len;

		_e2_file_info_dialog_get_file_type (localpath, file_type,
			       sizeof (file_type));
		if ((len = strlen (file_type)) > (FILE_TYPE_LENGTH - 2))
		{
			for (len = FILE_TYPE_LENGTH - 4; len < FILE_TYPE_LENGTH-1; len++)
			{
				file_type[len] = '.';
			}
			file_type[FILE_TYPE_LENGTH-1] = '\0';
		}
		else
		{
			file_type[len-1] = '\0';
		}
		s2 = e2_utf8_from_locale (file_type);	//freeme later
	}
//	else if (thisis_dir)
	else if (e2_fs_is_dir3 (localpath))
		s2 =  _("Directory");
	else if (S_ISCHR (statbuf.st_mode))
		s2 = _("Character Device");
	else if (S_ISBLK (statbuf.st_mode))
		s2 = _("Block Device");
	else if (S_ISFIFO (statbuf.st_mode))
		s2 = _("FIFO Pipe");
	else if (S_ISSOCK (statbuf.st_mode))
		s2 = _("Socket");
	else
		s2 = _("Unknown");

	tmsg = g_strdup_printf ("%s: %s", s1, s2);
	e2_widget_add_label (dialog_vbox, tmsg, 0, TRUE, 0);
	g_free (tmsg);
	if (S_ISREG (statbuf.st_mode))
		g_free (s2);
	if (S_ISLNK (statbuf.st_mode))
	{
		gchar linkto[PATH_MAX+NAME_MAX];
		gchar *filename = g_strdup (localpath);
		gint len = strlen (filename);
		if (filename[len-1] == G_DIR_SEPARATOR)
			filename[len-1] = '\0';

		len = readlink (filename, linkto, sizeof (linkto));	//FIXME vfs
		if (len > 0)
		{
			linkto[len] = '\0';
			gchar *utf = F_DISPLAYNAME_FROM_LOCALE (linkto);
			g_string_printf (label_text, " %s %s", _("to"), utf);
			F_FREE (utf);
			//check the real target, without looking thru chained links
			struct stat statbuf2;	//don't disturb the size etc data for the link itelf
			if (e2_fs_lstat (linkto, &statbuf2) && errno == ENOENT)
				g_string_append_printf (label_text, "\n %s",_("(which is <b>missing</b>)"));

			e2_widget_add_label (dialog_vbox, label_text->str, 0, TRUE, 0);
		}
		else
			e2_widget_add_label (dialog_vbox, _("(Cannot resolve the link)"), 0, TRUE, 0);

		g_free (filename);
	}

	table = e2_widget_add_table (dialog_vbox, 9, 2, FALSE, TRUE, E2_PADDING);
	gtk_table_set_col_spacings (GTK_TABLE (table), E2_PADDING);
	
	e2_widget_add_label_to_table (table, _("Size:"), 0, 0,1,0,1);

	if (statbuf.st_size < 10240) // less than 10K
		g_string_printf (label_text, "%d %s", (gint) statbuf.st_size, _("bytes"));
	else if (statbuf.st_size < 1048576) // less than a meg
		g_string_printf (label_text, "%.2f %s", (gfloat) statbuf.st_size / 1024.0, _("kbytes"));
	else // more than a meg
		g_string_printf (label_text, "%.2f %s", (gfloat) statbuf.st_size / 1048576.0, _("Mbytes"));

	e2_widget_add_label_to_table (table, label_text->str, 0, 1,2,0,1);

	GtkWidget *sep = gtk_hseparator_new ();
	gtk_table_attach_defaults (GTK_TABLE (table), sep, 0,2,1,2);

	e2_widget_add_label_to_table (table, _("User:"), 0, 0,1,2,3);
	if ((pw_buf = getpwuid (statbuf.st_uid)) != NULL)
		g_string_printf (label_text, "%s", pw_buf->pw_name);
	else
		g_string_printf (label_text, "%d", (guint) statbuf.st_uid);
	e2_widget_add_label_to_table (table, label_text->str, 0, 1,2,2,3);

	e2_widget_add_label_to_table (table, _("Group:"), 0, 0,1,3,4);
	if ((grp_buf = getgrgid (statbuf.st_gid)) != NULL)
		g_string_printf (label_text, "%s", grp_buf->gr_name);
	else
		g_string_printf (label_text, "%d", (guint) statbuf.st_gid);
	e2_widget_add_label_to_table (table, label_text->str, 0, 1,2,3,4);
	e2_fs_get_perm_string (perm_string, sizeof (perm_string), statbuf.st_mode);
	g_string_assign (label_text, perm_string);
	g_string_insert_c (label_text, 7, ' ');
	g_string_insert_c (label_text, 4, ' ');
	g_string_insert_c (label_text, 1, ' ');
	e2_widget_add_label_to_table (table, _("Permissions:"), 0, 0,1,4,5);
	e2_widget_add_label_to_table (table, label_text->str, 0, 1,2,4,5);

	sep = gtk_hseparator_new ();
	gtk_table_attach_defaults (GTK_TABLE (table), sep, 0,2,5,6);

	tm_ptr = localtime (&statbuf.st_atime);
	strftime (date_string, sizeof (date_string), "%a %b %d %I:%M %p", tm_ptr);
	e2_widget_add_label_to_table (table, _("Accessed:"), 0, 0,1,6,7);
	e2_widget_add_label_to_table (table, date_string, 0, 1,2,6,7);

	tm_ptr = localtime (&statbuf.st_mtime);
	strftime (date_string, sizeof (date_string), "%a %b %d %I:%M %p", tm_ptr);
	e2_widget_add_label_to_table (table, _("Modified:"), 0, 0,1,7,8);
	e2_widget_add_label_to_table (table, date_string, 0, 1,2,7,8);

	tm_ptr = localtime (&statbuf.st_ctime);
	strftime (date_string, sizeof (date_string), "%a %b %d %I:%M %p", tm_ptr);
	e2_widget_add_label_to_table (table, _("Changed:"), 0, 0,1,8,9);
	e2_widget_add_label_to_table (table, date_string, 0, 1,2,8,9);

	g_string_free (label_text,TRUE);

	if (multi)
	{
		E2_BUTTON_NOTOALL.showflags &= ~!E2_BTN_DEFAULT;
		e2_dialog_add_defined_button (dialog, &E2_BUTTON_NOTOALL);
		e2_dialog_set_negative_response (dialog, E2_RESPONSE_NOTOALL);
	}
	E2_BUTTON_CLOSE.showflags |= E2_BTN_DEFAULT;
	e2_dialog_add_defined_button (dialog, &E2_BUTTON_CLOSE);

	e2_dialog_setup (dialog, app.main_window);
	
#ifdef E2_ASYNC
	gdk_threads_enter ();
#endif
	e2_dialog_run (dialog, app.main_window, 0);
	gtk_main ();
#ifdef E2_ASYNC
	gdk_threads_leave ();
#endif
	return choice;
}
